#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

#include <glib.h>
#include <gst/gst.h>

#include "util.h"

static gboolean
clpi_dump_clpi_entry (guint8 * buf, guint8 * end, guint32 start_offset)
{
  CHECK_AVAIL_DECL;
  guint32 table_size;
  guint16 unknown1;
  guint8 stream_type;
  guint8 app_type;
  guint32 unknown2;
  gboolean is_atc_delta;
  guint32 ts_recording_rate;
  guint32 num_src_packets;
  guint16 type_info_size;

  CHECK_AVAIL (buf, end, 20, need_data_clpi_entry);
  table_size = GST_READ_UINT32_BE (buf);
  buf += 4;
  g_print ("\n=== Clip Info, size %u bytes (offset 0x%08x-0x%08x) ===\n",
      table_size, start_offset, start_offset+table_size);
  if (end > buf + table_size)
    end = buf + table_size;

  unknown1 = GST_READ_UINT16_BE (buf);
  stream_type = buf[2];
  app_type = buf[3];
  unknown2 = GST_READ_UINT32_BE (buf + 4);
  is_atc_delta = (unknown2 & 0x01) ? TRUE : FALSE;
  unknown2 >>= 1;

  ts_recording_rate = GST_READ_UINT32_BE (buf + 8);
  num_src_packets = GST_READ_UINT32_BE (buf + 12);

  g_print ("Unknown byte 1: %d\n", unknown1);
  g_print ("Stream type: 0x%02x\n", stream_type);
  g_print ("Application Type: 0x%02x\n", app_type);
  g_print ("unknown 31 bits: 0x%08x\n", unknown2);
  g_print ("is ATC delta: %d\n", is_atc_delta);
  g_print ("TS recording rate: %d\n", ts_recording_rate);
  g_print ("num_src_packets: %d\n", num_src_packets);

  buf += 16;

  CHECK_AVAIL (buf, end, 128, need_data_clpi_table);
  g_print ("Skipping 128 reserved bytes\n");
  buf += 128;

  CHECK_AVAIL (buf, end, 2, need_data_typeinfo);

  type_info_size = GST_READ_UINT16_BE (buf);
  buf += 2;

  CHECK_AVAIL (buf, end, type_info_size, need_data_typeinfo);
  g_print ("%d bytes of type info:\n", type_info_size);
  dump_bytes (buf, end - buf, 4);
  buf += type_info_size;

  if (is_atc_delta) {
    gint i;
    guint8 rsvd1;
    guint8 atc_delta_cnt;

    /* Parse ATC delta table */
    CHECK_AVAIL (buf, end, 2, need_data_atc_delta);
    rsvd1 = buf[0];
    atc_delta_cnt = buf[1];
    buf += 2;

    g_print ("ATC Delta table\n");
    g_print ("Unknown byte: 0x%02x\n", rsvd1);
    g_print ("ATC delta count: %d\n", atc_delta_cnt);

    CHECK_AVAIL (buf, end, atc_delta_cnt * 14, need_data_atc_delta);
    for (i = 0; i < atc_delta_cnt; i++) {
      g_print ("  ATC delta %d:\n", i);
      dump_bytes (buf, 14, 4);
      buf += 14;
    }
  }

  return TRUE;
need_data_clpi_entry:
  CHECK_PRINT_ERROR ("Failed to read Clip Info header");
  return FALSE;
need_data_clpi_table:
  CHECK_PRINT_ERROR ("Failed to read Clip Info 128 byte table");
  return FALSE;
need_data_typeinfo:
  CHECK_PRINT_ERROR ("Failed to read Clip type info");
  return FALSE;
need_data_atc_delta:
  CHECK_PRINT_ERROR ("Failed to read Clip ATC delta table");
  return FALSE;
}

static gboolean
clpi_dump_sequence_table (guint8 * buf, guint8 * end, guint32 start_offset)
{
  CHECK_AVAIL_DECL;
  guint32 table_size;
  guint8 unknown1, atc_seq_n;
  guint32 atc_offset;
  guint8 stc_entries_n;
  guint8 stc_id;
  gint i;

  CHECK_AVAIL (buf, end, 12, need_data_seq_table);
  table_size = GST_READ_UINT32_BE (buf);
  buf += 4;
  g_print ("\n=== Sequence Table, size %u bytes (offset 0x%08x-0x%08x) ===\n",
      table_size, start_offset, start_offset+table_size);
  if (end > buf + table_size)
    end = buf + table_size;

  unknown1 = buf[0];
  atc_seq_n = buf[1];
  atc_offset = GST_READ_UINT32_BE (buf + 2);
  stc_entries_n = buf[6];
  stc_id = buf[7];
  buf += 8;

  g_print ("Unknown byte 1: %d\n", unknown1);
  if (atc_seq_n != 1) {
    g_print
        ("!!! ATC sequence count (%d) != 1, please submit file for analysis\n",
        atc_seq_n);
  }
  if (atc_offset != 0) {
    g_print ("!!! ATC offset (%d) != 0, please submit file for analysis\n",
        atc_offset);
  }
  g_print ("ATC sequence count %d offset %d\n", atc_seq_n, atc_offset);
  g_print ("STC entry count %d, ID %d\n", stc_entries_n, stc_id);

  for (i = 0; i < stc_entries_n; i++) {
    guint16 pcr_pid;            /* PID of the PCR stream */
    guint32 packet_no;          /* Packet number in the mpegts to set STC */
    guint64 stc_start;          /* First presentation time */
    guint64 stc_end;            /* End presentation time */

    CHECK_AVAIL (buf, end, 14, need_data_seq_stc);

    pcr_pid = GST_READ_UINT16_BE (buf);
    packet_no = GST_READ_UINT32_BE (buf + 2);
    stc_start = GST_READ_UINT32_BE (buf + 6) << 1;
    stc_end = GST_READ_UINT32_BE (buf + 10) << 1;

    g_print ("  STC %d: PCR PID 0x%04x packet %d\n"
        "        start %" G_GUINT64_FORMAT " end %" G_GUINT64_FORMAT "\n",
        i, pcr_pid, packet_no, stc_start, stc_end);
    buf += 14;
  }

  return TRUE;
need_data_seq_table:
  CHECK_PRINT_ERROR ("Failed to read Sequence Table header");
  return FALSE;
need_data_seq_stc:
  CHECK_PRINT_ERROR ("Failed to read STC entry");
  return FALSE;
}

static gboolean
clpi_dump_program_table (guint8 * buf, guint8 * end, guint32 start_offset)
{
  CHECK_AVAIL_DECL;
  guint32 table_size;
  guint8 unknown1;              /* Unknown. ID? */
  guint8 pgm_entries_n;         /* No. of program entries. */
  guint32 pgm_seq_start;        /* Start of the program sequence (must be 0?) */
  guint16 pmt_pid;              /* PID of the PMT stream */
  guint8 stream_entries_n;      /* Number of stream entries in table */
  guint8 group_cnt;             /* ??? */
  gint i;

  CHECK_AVAIL (buf, end, 14, need_data_pgm_table);

  table_size = GST_READ_UINT32_BE (buf);
  buf += 4;
  g_print ("\n=== Program Table, size %u bytes (offset 0x%08x-0x%08x) ===\n",
      table_size, start_offset, start_offset+table_size);
  if (end > buf + table_size)
    end = buf + table_size;

  CHECK_AVAIL (buf, end, table_size, need_data_pgm_table);

  unknown1 = buf[0];
  pgm_entries_n = buf[1];

  pgm_seq_start = GST_READ_UINT32_BE (buf + 2);
  pmt_pid = GST_READ_UINT16_BE (buf + 6);
  stream_entries_n = buf[8];
  group_cnt = buf[9];

  buf += 10;

  if (pgm_entries_n != 1) {
    g_print ("!!! program table entry count (%d) != 1 -"
        " please submit file for analysis\n", pgm_entries_n);
  }

  g_print ("Unknown byte 1: %d\n", unknown1);
  g_print ("Program entries: %d sequence start: %d\n",
      pgm_entries_n, pgm_seq_start);
  g_print ("PMT PID: 0x%04x stream entries: %d group count: %d\n",
      pmt_pid, stream_entries_n, group_cnt);

  for (i = 0; i < stream_entries_n; i++) {
    guint16 pid;                /* elementary PID */
    guint8 len;                 /* Length of table entry */
    guint8 type;                /* Type of stream? */
    guint8 desc_len;
    CHECK_AVAIL (buf, end, 4, need_data_pgm_table);

    pid = GST_READ_UINT16_BE (buf);
    len = buf[2];
    g_print ("  PID 0x%04x has %d bytes data\n", pid, len);
    buf += 3;

    CHECK_AVAIL (buf, end, len, need_data_pgm_table);

    if (len > 0) {
      type = buf[0];
      g_print ("    Type byte 0x%x, data:\n", type);
      dump_bytes (buf + 1, len - 1, 4);
      buf += len;
    }
  }

  if (buf < end) {
    g_print ("%u unused bytes in stream entry header:\n", end-buf);
    dump_bytes (buf, end - buf, 0);
  }

  return TRUE;
need_data_pgm_table:
  CHECK_PRINT_ERROR ("Failed to read Program Table header");
  return FALSE;
}

static gboolean
clpi_dump_ep_fine_table (guint8 * buf, guint8 * end, guint32 first,
    guint32 last, guint64 PTS_msb, guint32 packet_no_msb)
{
  CHECK_AVAIL_DECL;
  static guint64 last_pts;
  static guint32 last_packet;

  /* Check the table entries are within the data */
  CHECK_AVAIL (buf, end, last*4, need_data_fine_entries);

  end = buf + (last * 4);
  buf += (first * 4);

  PTS_msb &= G_GUINT64_CONSTANT (0x1FFF00000);  /* Mask top 13 bits */
  packet_no_msb &= 0xFFFE0000;  /* Mask top 15 bits */

  while (buf < end) {
    gboolean angle_switch;
    guint8 unknown;
    guint64 pts;
    guint32 packet_no;

    angle_switch = !!(buf[0] & 0x80);
    unknown = (buf[0] & 0x70) >> 4;
    pts = PTS_msb | ((((buf[0] << 8) | buf[1]) << 8) & 0xfff00);
    packet_no =
        packet_no_msb | (((buf[1] & 0x1) << 16) | (buf[2] << 8) | buf[3]);
    if (first != 0) {
      g_print ("    fine %d: PTS %" G_GUINT64_FORMAT " (diff %" G_GUINT64_FORMAT
          ") packet %u (diff %d) unknown %d "
          "angle switch %d\n", first, pts, pts - last_pts, packet_no,
          packet_no - last_packet, unknown, angle_switch);
    } else {
      g_print ("    fine %d: PTS %" G_GUINT64_FORMAT " packet %u unknown %d "
          "angle switch %d\n", first, pts, packet_no, unknown, angle_switch);
    }
    buf += 4;
    first++;

    last_pts = pts;
    last_packet = packet_no;
  }

  return TRUE;
need_data_fine_entries:
  CHECK_PRINT_ERROR ("Failed to read Fine seek table entries");
  return FALSE;
}

static gboolean
clpi_dump_ep_map_table (guint8 * buf, guint8 * end, guint16 coarse_n,
    guint32 fine_n)
{
  CHECK_AVAIL_DECL;
  guint8 *ep_map_table_start;
  guint8 *fine_table;
  guint32 fine_table_offset;
  guint32 fine_idx_entry_prev = 0;
  guint64 PTS_msb_prev = 0;
  guint32 packet_no_msb_prev = 0;
  gint i;

  CHECK_AVAIL (buf, end, 8*coarse_n + 4, need_data_ep_map);

  ep_map_table_start = buf;

  fine_table_offset = GST_READ_UINT32_BE (buf);
  fine_table = ep_map_table_start + fine_table_offset;
  g_print ("  fine table offset: %u\n", fine_table_offset);
  buf += 4;

  for (i = 0; i < coarse_n; i++) {
    guint32 fine_idx_entry;
    guint64 PTS_msb;
    guint32 packet_no_msb;

    fine_idx_entry = (buf[0] << 10) | (buf[1] << 2) | (buf[2] >> 6);
    PTS_msb = (((guint64) (buf[2]) << 8) | buf[3]) << 19;
    PTS_msb &= G_GUINT64_CONSTANT (0x1FFF80000);
    packet_no_msb = GST_READ_UINT32_BE (buf + 4);

    if (fine_idx_entry > fine_idx_entry_prev) {
      /* Dump fine table region */
      clpi_dump_ep_fine_table (fine_table, end,
          fine_idx_entry_prev, fine_idx_entry, PTS_msb_prev,
          packet_no_msb_prev);
    }

    g_print ("  entry %4d: Fine idx %u PTS msb %" G_GUINT64_FORMAT
        " packet msb %u\n", i, fine_idx_entry, PTS_msb, packet_no_msb);

    fine_idx_entry_prev = fine_idx_entry;
    PTS_msb_prev = PTS_msb;
    packet_no_msb_prev = packet_no_msb;

    buf += 8;
  }

  if (fine_n > fine_idx_entry_prev) {
    /* Dump fine table region */
    clpi_dump_ep_fine_table (fine_table, end, fine_idx_entry_prev, fine_n,
        PTS_msb_prev, packet_no_msb_prev);
  }
  return TRUE;
need_data_ep_map:
  CHECK_PRINT_ERROR ("Failed to read Entry Point Map table entries");
  return FALSE;
}

static gboolean
clpi_dump_entrypoint_map (guint8 * buf, guint8 * end)
{
  CHECK_AVAIL_DECL;
  guint8 *ep_map_start;
  guint8 unknown1;
  guint8 epmap_pids_n;          /* Number of PID maps in the table */
  gint i;

  CHECK_AVAIL (buf, end, 2, need_data_ep_map);
  ep_map_start = buf;

  unknown1 = buf[0];
  epmap_pids_n = buf[1];
  buf += 2;

  g_print ("Entry point map has unknown field %d (0x%x) and %d ES pids\n",
      unknown1, unknown1, epmap_pids_n);

  for (i = 0; i < epmap_pids_n; i++) {
    guint16 pmt_pid;
    guint16 epmap_unknown;
    guint8 type;
    guint16 coarse_entries_n;
    guint32 fine_entries_n;
    guint32 coarse_table_offset;
    guint64 tmp;

    if (buf + 12 > end) {
      g_print ("Failure parsing entry point map: seek table entry truncated\n");
      return FALSE;
    }

    pmt_pid = GST_READ_UINT16_BE (buf);

    tmp = GST_READ_UINT64_BE (buf);
    /* Take 10 bits, unknown data: */
    epmap_unknown = (buf[2] << 2) | (buf[3] >> 6);
    /* Take 4 bits: table type, or ID? */
    type = (buf[3] >> 2) & 0xf;
    /* Take 16 bits: No of coarse entries */
    coarse_entries_n = ((guint16) buf[3] << 14) | ((guint16) buf[4] << 6) |
        (((guint16) buf[5] >> 2) & 0x3f);
    /* Take 18 bits, No of fine entries */
    fine_entries_n = ((guint16) (buf[5] & 0x3) << 16) | buf[6] << 8 | buf[7];
    /* Offset to the coarse table */
    coarse_table_offset = GST_READ_UINT32_BE (buf + 8);
    buf += 12;

    g_print ("PMT pid 0x%04x unknown field %d (0x%x) table type %d "
        "coarse entries: %d fine entries: %d.\n"
        "  Coarse table @ offset %d\n",
        pmt_pid, epmap_unknown, epmap_unknown, type, coarse_entries_n,
        fine_entries_n, coarse_table_offset);
    if (ep_map_start + coarse_table_offset > end) {
      g_print
          ("Error parsing entry point map: Coarse table offset too large\n");
      return FALSE;
    }
    clpi_dump_ep_map_table (ep_map_start + coarse_table_offset, end,
        coarse_entries_n, fine_entries_n);
  }

  return TRUE;
need_data_ep_map:
  CHECK_PRINT_ERROR ("Failed to read Entry Point Map table");
  return FALSE;
}

static gboolean
clpi_dump_cpi_table (guint8 * buf, guint8 * end)
{
  guint32 cpi_table_size;
  guint8 cpi_table_unknown;
  guint8 cpi_table_type;
  gint i;

  if (buf + 6 > end) {
    g_print ("Failure parsing CPI table: header truncated\n");
    return FALSE;
  }
  cpi_table_size = GST_READ_UINT32_BE (buf);

  cpi_table_unknown = GST_READ_UINT16_BE (buf + 4);
  cpi_table_type = cpi_table_unknown & 0x0f;
  cpi_table_unknown >>= 4;

  if (buf + cpi_table_size > end) {
    g_print ("Failure parsing CPI table: not enough bytes for size %d\n",
        cpi_table_size);
    return FALSE;
  }

  g_print ("\n\n=== CPI (seek) Table ===\n");
  /* dump_bytes (buf, cpi_table_size, 2); */

  g_print ("CPI table length: %d. Unknown field 1: 0x%x, table type %d\n",
      cpi_table_size, cpi_table_unknown, cpi_table_type);

  if (cpi_table_type != 1)
    g_print ("!!! CPI table type (%d) != 1, please submit file for analysis\n",
        cpi_table_type);

  buf += 6;

  return clpi_dump_entrypoint_map (buf, end);
}

static gboolean
dump_clpi (guint8 * buf, guint32 len)
{
  guint8 *pos, *end;
  guint32 file_id;
  guint32 file_ver;
  guint32 clpi_entry_offset;
  guint32 sequence_table_offset;
  guint32 program_table_offset;
  guint32 cpi_table_offset;
  guint32 mark_table_offset;
  guint32 extension_table_offset;
  const guint32 bd_version = GST_READ_UINT32_BE ("0200");
  const guint32 avchd_version = GST_READ_UINT32_BE ("0100");

  pos = buf;
  end = pos + len;

  /* Need at least 28 bytes */
  if (pos + 28 > end) {
    g_print ("Not enough bytes to be a CLPI file\n");
    return FALSE;
  }

  file_id = GST_READ_UINT32_BE (pos);
  file_ver = GST_READ_UINT32_BE (pos + 4);
  pos += 8;

  if (file_id != GST_READ_UINT32_BE ("HDMV")) {
    g_print ("Invalid file ID: %" GST_FOURCC_FORMAT ". Should be HDMV\n",
        (guint8)(file_id >> 24), (guint8)(file_id >> 16),
        (guint8)(file_id >> 8), (guint8)(file_id));
    return FALSE;
  }
  if (file_ver != bd_version && file_ver != avchd_version) {
    g_print ("Previously unseen file version: %" GST_FOURCC_FORMAT
        ". Please send it in for analysis\n", GST_FOURCC_ARGS (file_ver));
    return FALSE;
  }

  /* CLPI entry always at byte 40 */
  clpi_entry_offset = 40;

  sequence_table_offset = GST_READ_UINT32_BE (pos);
  program_table_offset = GST_READ_UINT32_BE (pos + 4);
  cpi_table_offset = GST_READ_UINT32_BE (pos + 8);
  mark_table_offset = GST_READ_UINT32_BE (pos + 12);
  extension_table_offset = GST_READ_UINT32_BE (pos + 16);
  pos += 20;

  g_print ("Read file header. Sequence Table @ 0x%08x Program Table @ 0x%08x\n",
      sequence_table_offset, program_table_offset);
  g_print
      ("    CPI table @ 0x%08x Mark table @ 0x%08x Extension table @ 0x%08x\n",
      cpi_table_offset, mark_table_offset, extension_table_offset);

  if (clpi_entry_offset >= len) {
    g_print
        ("Not enough data (%d bytes), or CLPI entry offset offset (%d) invalid\n",
        len, clpi_entry_offset);
    return FALSE;
  }
  if (!clpi_dump_clpi_entry (buf + clpi_entry_offset, end,
      clpi_entry_offset))
    return FALSE;

  if (sequence_table_offset >= len) {
    g_print
        ("Not enough data (%d bytes), or sequence table offset (%d) invalid\n",
        len, sequence_table_offset);
    return FALSE;
  }
  if (!clpi_dump_sequence_table (buf + sequence_table_offset, end,
      sequence_table_offset))
    return FALSE;

  if (program_table_offset >= len) {
    g_print
        ("Not enough data (%d bytes), or program table offset (%d) invalid\n",
        len, program_table_offset);
    return FALSE;
  }
  if (!clpi_dump_program_table (buf + program_table_offset, end,
      program_table_offset))
    return FALSE;

  if (cpi_table_offset >= len) {
    g_print ("Not enough data (%d bytes), or CPI table offset (%d) invalid\n",
        len, cpi_table_offset);
    return FALSE;
  }
  if (!clpi_dump_cpi_table (buf + cpi_table_offset, end))
    return FALSE;

  return TRUE;
}

int
main (int argc, char **argv)
{
  int fd, eof = 0;
  int ret = 0;
  guint8 *buf = NULL;
  gchar *contents;
  gsize size;
  GError *error = NULL;
  guint32 avail = 0;

  if (argc < 2) {
    g_print ("Usage: %s <filename>\n", argv[0]);
    return 1;
  }

  g_file_get_contents (argv[1], &contents, &size, &error);
  if (error != NULL) {
    g_print ("Failed opening file %s. Error %s\n", argv[1], error->message);
    g_error_free (error);
    return 1;
  }

  buf = (guint8 *) contents;
  dump_clpi (buf, size);

done:
  g_free (contents);
  close (fd);

  return (ret < 0) ? ret : 0;
}
