#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

#include <glib.h>
#include <gst/gst.h>

#include "util.h"

const gchar *still_types[] = {
  "No Still", "Limited Still", "Unlimited Still"
};

const guint n_still_types = G_N_ELEMENTS (still_types);

typedef enum _MplsStreamType
{
  MPLS_ST_VIDEO = 0,
  MPLS_ST_AUDIO,
  MPLS_ST_PG,
  MPLS_ST_IG,
  MPLS_ST_AUDIO_2,
  MPLS_ST_VIDEO_2,
  MPLS_ST_PIP_PG,

  /* Unknown stream type entries: (future expansion?) */
  MPLS_ST_UNKNOWN_1,
  MPLS_ST_UNKNOWN_2,
  MPLS_ST_UNKNOWN_3,
  MPLS_ST_UNKNOWN_4,
  MPLS_ST_UNKNOWN_5,

  MPLS_ST_MAX
} MplsStreamType;

static gchar *mpls_stream_type_names[] = {
  "Video", "Audio", "Presentation Graphics", "Interactive Graphics",
  "Audio_2", "Video_2", "Picture-in-Picture Presentation Graphics",
  "Unknown 1", "Unknown 2", "Unknown 3", "Unknown 4", "Unknown 5"
};

static const gchar *
stream_entry_type_to_str (guint8 stream_entry_type)
{
  switch (stream_entry_type) {
    case 0x1:
      return "Play Item";
    case 0x2:
      return "Sub Path";
    case 0x3:
      return "In Mux Sub Path";
    default:
      return "Unknown";
  }
}

static gboolean
mpls_dump_stream_entry_hdr (guint8 * buf, guint8 * end, gint i)
{
  CHECK_AVAIL_DECL;
  guint8 stream_entry_size = (end - buf);
  guint8 stream_entry_type;

  CHECK_AVAIL (buf, end, 1, need_data);

  stream_entry_type = buf[0];
  buf += 1;

  g_print ("%8sStream entry %d (%u bytes) type %u (%s)\n", " ", i,
      stream_entry_size, stream_entry_type,
      stream_entry_type_to_str (stream_entry_type));
  switch (stream_entry_type) {
    case 0x1:{
      guint16 stream_pid;
      CHECK_AVAIL (buf, end, 2, need_data);
      stream_pid = GST_READ_UINT16_BE (buf);
      g_print ("%10sPID: 0x%04x\n", " ", stream_pid);
      buf += 2;
      break;
    }
    case 0x2:{
      guint8 path_id;
      guint8 sub_clip_id;
      guint16 stream_pid;
      CHECK_AVAIL (buf, end, 4, need_data);
      path_id = buf[0];
      sub_clip_id = buf[1];
      stream_pid = GST_READ_UINT16_BE (buf + 2);
      g_print ("%10sPath Id %u Sub-clip Id %u PID: 0x%04x\n", " ",
          path_id, sub_clip_id, stream_pid);
      buf += 4;
      break;
    }
    case 0x3:{
      guint16 stream_pid;
      guint8 sub_path_id;
      CHECK_AVAIL (buf, end, 3, need_data);
      sub_path_id = *buf++;
      stream_pid = GST_READ_UINT16_BE (buf);
      buf += 2;
      g_print ("%10sSub-path Id %u PID: 0x%04x\n", " ",
          sub_path_id, stream_pid);
      break;
    }
    default:
      dump_bytes (buf, stream_entry_size, 10);
      break;
  }

  if (buf < end) {
    g_print ("%10s%u unused bytes in stream entry header:\n", "", end - buf);
    dump_bytes (buf, end - buf, 12);
  }
  return TRUE;
need_data:
  CHECK_PRINT_ERROR ("Failed to read Stream Entry header");
  return FALSE;
}

static gboolean
mpls_dump_stream_attrs (guint8 * buf, guint8 * end)
{
  CHECK_AVAIL_DECL;
  const gchar *stream_type_name = "unknown";
  guint8 stream_attrs_size = (end - buf);
  guint8 stream_type;

  CHECK_AVAIL (buf, end, 1, need_data);
  stream_type = buf[0];
  buf += 1;

  if (stream_type == 0x02)
    stream_type_name = "MPEG";
  else if (stream_type == 0x1b)
    stream_type_name = "AVC";
  else if (stream_type == 0xea)
    stream_type_name = "VC-1";
  else if (stream_type >= 0x80 && stream_type <= 0x8f)
    stream_type_name = "AC-3";
  else if (stream_type >= 0xa0 && stream_type <= 0xaf)
    stream_type_name = "EAC-3";
  else if (stream_type == 0x90)
    stream_type_name = "PG";
  else if (stream_type == 0x91)
    stream_type_name = "IG";
  else if (stream_type == 0x92)
    stream_type_name = "Text";

  g_print ("%8sStream attributes (%u bytes) stream type 0x%02x (%s):\n", "",
      stream_attrs_size, stream_type, stream_type_name);

  if ((stream_type >= 0x80 && stream_type <= 0x8f) ||
      (stream_type >= 0xa0 && stream_type <= 0xaf)) {
    /* Audio stream */
    guint8 unknown1, unknown2;
    gchar lang_code[4];
    CHECK_AVAIL (buf, end, 4, need_data);

    unknown1 = buf[0] >> 4;
    unknown2 = buf[0] & 0x0f;
    buf++;
    memcpy (lang_code, buf, 3);
    lang_code[3] = '\0';
    buf += 3;

    g_print ("%10sUnknown 4-bit fields: %d and %d, lang code %s\n", "",
        unknown1, unknown2, lang_code);
  } else if (stream_type == 0x90 || stream_type == 0x91) {
    /* PG or IG stream */
    gchar lang_code[4];
    CHECK_AVAIL (buf, end, 3, need_data);
    memcpy (lang_code, buf, 3);
    lang_code[3] = '\0';
    buf += 3;
    g_print ("%10slang code %s\n", "", lang_code);
  } else if (stream_type == 0x92) {
    /* Text stream */
    guint8 charset;
    gchar lang_code[4];
    CHECK_AVAIL (buf, end, 4, need_data);
    charset = *buf++;
    memcpy (lang_code, buf, 3);
    lang_code[3] = '\0';
    buf += 3;
    g_print ("%10scharset 0x%x lang code %s\n", "", charset, lang_code);
  }

  if (buf < end) {
    g_print ("%10s%u unused bytes in stream attributes:\n", "", end - buf);
    dump_bytes (buf, end - buf, 12);
  }
  return TRUE;
need_data:
  CHECK_PRINT_ERROR ("Failed to read Stream attributes");
  return FALSE;
}

static gboolean
mpls_dump_stn_table (guint8 * buf, guint8 * end)
{
  CHECK_AVAIL_DECL;
  guint16 unknown;
  guint n_streams[MPLS_ST_MAX];
  gint st, i;

  CHECK_AVAIL (buf, end, 2 + 2 + 7, need_data_stn);

  /* Skip 16-bit size, already read */
  buf += 2;
  unknown = GST_READ_UINT16_BE (buf);
  buf += 2;
  for (st = MPLS_ST_VIDEO; st < MPLS_ST_MAX; st++)
    n_streams[st] = *buf++;
  g_print ("    Unknown 16-bit field: %u\n", unknown);
  g_print ("    Video streams: %u, Audio streams: %u\n",
      n_streams[MPLS_ST_VIDEO], n_streams[MPLS_ST_AUDIO]);
  g_print ("    Presentation Graphics streams: %u, "
      "Interactive Graphics streams: %u\n", n_streams[MPLS_ST_PG],
      n_streams[MPLS_ST_IG]);
  g_print ("    2nd-ary video streams: %u, 2nd-ary audio streams: %u\n",
      n_streams[MPLS_ST_VIDEO_2], n_streams[MPLS_ST_AUDIO_2]);
  g_print ("    Picture-in-Picture Presentation Graphics streams: %u\n",
      n_streams[MPLS_ST_PIP_PG]);

  /* Dump stream entries with known types */
  for (st = 0; st < MPLS_ST_MAX /* MPLS_ST_AUDIO_2 */ ; st++) {
    for (i = 0; i < n_streams[st]; i++) {
      guint8 blk_size;
      g_print ("%6s%s stream %u:\n", " ", mpls_stream_type_names[st], i);

      if (st >= MPLS_ST_AUDIO_2) {
        g_print ("%6sBytes remaining:\n", "");
        dump_bytes (buf, end - buf, 8);
      }

      /* Check 1 byte avail to read size: */
      CHECK_AVAIL (buf, end, 1, need_data_stream_entry);
      blk_size = *buf++;
      CHECK_AVAIL (buf, end, blk_size, need_data_stream_entry);

      mpls_dump_stream_entry_hdr (buf, buf + blk_size, i);
      buf += blk_size;

      /* Now read stream attributes for this stream */
      CHECK_AVAIL (buf, end, 1, need_data_stream_attr);
      blk_size = buf[0];
      buf += 1;

      CHECK_AVAIL (buf, end, blk_size, need_data_stream_attr);

      mpls_dump_stream_attrs (buf, buf + blk_size);
      buf += blk_size;

      /* This seems true for the samples I have (King Kong, Fringe Season 1,
       * Wall-E), there are extra bytes after the stream attributes */
      if (st == MPLS_ST_AUDIO_2 || st == MPLS_ST_VIDEO_2) {
        guint8 *tmp = buf;
        blk_size = 0;
        while (tmp < end) {
          if (tmp[0] == 0)
            break;
          blk_size += 2;
          tmp += 2;
        }
        if (st == MPLS_ST_AUDIO_2)
          blk_size += 2;        /* Skip 2 extra unknown bytes at the end */
        else
          blk_size += 4;        /* Skip 4 extra unknown bytes at the end */
        g_print ("%8sUnknown stream info bytes:\n", "");
        dump_bytes (buf, blk_size, 10);
        buf += blk_size;
      }
    }
  }

  /* Dump the bytes for any unknown streams */
  if (buf != end) {
    g_print ("    *** %u unused bytes remain in STN table:\n", end - buf);
    dump_bytes (buf, end - buf, 4);
  }
  return TRUE;
need_data_stn:
  CHECK_PRINT_ERROR ("Failed to parse STN table");
  return FALSE;
need_data_stream_entry:
  CHECK_PRINT_ERROR ("Failed to parse Stream Entry header");
  return FALSE;
need_data_stream_attr:
  CHECK_PRINT_ERROR ("Failed to parse Stream attributes");
  return FALSE;
}

static gboolean
mpls_dump_play_item (guint8 * buf, guint8 * end)
{
  CHECK_AVAIL_DECL;
  guint8 file_id[6];
  guint32 unknown;
  guint16 unknown2;
  gboolean angles_flag;
  guint8 connection;
  guint8 stc_id;
  guint32 stc_start, stc_end;
  guint64 unknown3;
  gboolean unknown4_flag;
  guint8 unknown5;
  guint8 still_type;
  guint16 still_time;
  guint16 stn_table_size;

  CHECK_AVAIL (buf, end, 2 + 5 + 4 + 11 + 12, need_data_play_item);

  buf += 2;                     /* Skip size marker */
  /* File ID */
  memcpy (file_id, buf, 5);
  file_id[5] = '\0';
  buf += 5;

  unknown = GST_READ_UINT32_BE (buf);
  buf += 4;
  unknown2 = (buf[0] << 3) | (buf[1] >> 5);
  angles_flag = !!(buf[1] & 0x10);
  connection = buf[1] & 0x0f;
  stc_id = buf[2];
  stc_start = GST_READ_UINT32_BE (buf + 3);
  stc_end = GST_READ_UINT32_BE (buf + 7);
  buf += 11;

  g_print ("  File id: %s\n", file_id);
  g_print ("  Unknown 16-bit field: %u Angles flag %u\n",
      unknown2, angles_flag);
  g_print ("  Connection: %u STC id: %u\n", connection, stc_id);
  g_print ("  STC start: %u, STC end: %u\n", stc_start, stc_end);

  unknown3 = GST_READ_UINT64_BE (buf);
  unknown4_flag = !!(buf[8] & 0x80);
  unknown5 = buf[8] & 0x7f;

  still_type = buf[9];
  still_time = GST_READ_UINT16_BE (buf + 10);
  buf += 12;

  g_print ("  Unknown 64-bit field: %" G_GUINT64_FORMAT "\n", unknown3);
  g_print ("  Unknown flag %d. Unknown 7 bit field: %u\n",
      unknown4_flag, unknown5);

  g_print ("  Still type: %u (%s), Still time: %u\n", still_type,
      still_type < n_still_types ? still_types[still_type] : "Unknown",
      still_time);

  if (angles_flag) {
    guint8 n_angles, unknown6;
    gboolean angle_audio_flag;
    gboolean angle_seamless_flag;
    guint8 angle_file_id[6];
    guint32 unknown7;
    guint8 angle_stc_id;
    gint i;

    CHECK_AVAIL (buf, end, 2, need_data_play_item);

    n_angles = buf[0];
    unknown6 = buf[1] >> 2;
    angle_audio_flag = !!(buf[1] & 0x02);
    angle_seamless_flag = !!(buf[1] & 0x01);

    g_print ("  Multi-angle info: %u angles\n", n_angles);
    g_print ("  Unknown 6-bits: %u Audio change flag: %d Seamless flag %d\n",
        unknown6, angle_audio_flag, angle_seamless_flag);
    buf += 2;

    CHECK_AVAIL (buf, end, 10 * n_angles, need_data_play_item);

    for (i = 0; i < n_angles; i++) {
      memcpy (angle_file_id, buf, 5);
      angle_file_id[5] = '\0';
      buf += 5;

      unknown7 = GST_READ_UINT32_BE (buf);
      buf += 4;
      angle_stc_id = buf[0];
      buf += 1;
      g_print
          ("    angle %d: file id %s, 32-bit unknown field: %u, STC id: %u\n",
          i, angle_file_id, unknown7, angle_stc_id);
    }

  }

  /* STN table */
  stn_table_size = GST_READ_UINT16_BE (buf) + 2;
  CHECK_AVAIL (buf, end, stn_table_size, need_data_stn_table);
  g_print ("  STN table, size %u bytes:\n", stn_table_size);
  if (!mpls_dump_stn_table (buf, buf + stn_table_size))
    return FALSE;
  buf += stn_table_size;

  if (buf != end) {
    g_print ("  %u unused bytes remain:\n", end - buf);
    dump_bytes (buf, end - buf, 4);
  }

  return TRUE;
need_data_play_item:
  CHECK_PRINT_ERROR ("Failed to parse play item");
  return FALSE;
need_data_stn_table:
  CHECK_PRINT_ERROR ("Failed parsing STN table");
  return FALSE;
}

static gboolean
mpls_dump_subpath_item (guint8 * buf, guint8 * end)
{
  CHECK_AVAIL_DECL;
  guint8 unknown;
  guint8 sub_path_type;
  guint16 flags;
  guint8 unknown1;
  guint8 n_items;
  gint i;

  CHECK_AVAIL (buf, end, 6, need_data);
  unknown = *buf++;
  sub_path_type = *buf++;
  flags = GST_READ_UINT16_BE (buf);
  buf += 2;
  unknown1 = *buf++;
  n_items = *buf++;

  g_print ("%2sUnknown byte: %d (0x%x)\n", "", unknown, unknown);
  g_print ("%2sSub-path type 0x%x flags 0x%04x (%s%s)\n", "", sub_path_type,
      flags, flags & 0xfffe ? "unknown, " : "",
      flags & 0x01 ? "repeat" : "no repeat");
  g_print ("%2sUnknown byte: %d (0x%x)\n", "", unknown1, unknown1);
  g_print ("%2s%d items:\n", "", n_items);

  dump_bytes (buf, end - buf, 6);
  return TRUE;
need_data:
  CHECK_PRINT_ERROR ("Failed to parse sub-path item");
  return FALSE;
}

static gboolean
mpls_dump_play_items_table (guint8 * buf, guint8 * end, guint32 start_offset)
{
  CHECK_AVAIL_DECL;
  gint i;
  guint32 table_size;
  guint16 unknown;
  guint16 n_play_items;
  guint16 n_subpaths;

  CHECK_AVAIL (buf, end, 10, need_data_play_items);

  table_size = GST_READ_UINT32_BE (buf) + 4;
  unknown = GST_READ_UINT16_BE (buf + 4);
  n_play_items = GST_READ_UINT16_BE (buf + 6);
  n_subpaths = GST_READ_UINT16_BE (buf + 8);
  buf += 10;

  g_print ("=== Play Items table, size %u bytes (offset 0x%08x-0x%08x)\n",
      table_size, start_offset, start_offset + table_size);
  CHECK_AVAIL (buf, end, table_size, need_data_play_items);

  g_print ("Unknown 16 bit value: %u\n", unknown);
  g_print ("%u play items, %u sub paths\n", n_play_items, n_subpaths);

  for (i = 0; i < n_play_items; i++) {
    guint16 item_size = GST_READ_UINT16_BE (buf) + 2;
    CHECK_AVAIL (buf, end, item_size, need_data_play_items);
    g_print ("Play item %d: size %u bytes:\n", i + 1, item_size);
    if (!mpls_dump_play_item (buf, buf + item_size))
      return FALSE;
    buf += item_size;
  }
  for (i = 0; i < n_subpaths; i++) {
    guint32 item_size = GST_READ_UINT32_BE (buf) + 2;
    CHECK_AVAIL (buf, end, item_size, need_data_subpaths);
    g_print ("Sub-path item %d: size %u bytes:\n", i + 1, item_size);
    if (!mpls_dump_subpath_item (buf, buf + item_size))
      return FALSE;
    buf += item_size;
  }

  return TRUE;
need_data_play_table:
  CHECK_PRINT_ERROR ("Failed to parse play item table");
  return FALSE;
need_data_play_items:
  CHECK_PRINT_ERROR ("Failed to parse play item");
  return FALSE;
need_data_subpaths:
  CHECK_PRINT_ERROR ("Failed to parse sub-path table entry");
  return FALSE;
}

static gboolean
mpls_dump_mark_table (guint8 * buf, guint8 * end, guint32 start_offset)
{
  CHECK_AVAIL_DECL;
  guint32 table_size;
  guint16 n_entries;
  gint i;

  CHECK_AVAIL (buf, end, 6, need_data);
  table_size = GST_READ_UINT32_BE (buf);
  buf += 4;
  n_entries = GST_READ_UINT16_BE (buf);

  g_print ("=== Mark table, size %u bytes (offset 0x%08x-0x%08x) ===\n",
      table_size, start_offset, start_offset + table_size);
  g_print ("  %u table entries\n", n_entries);

  CHECK_AVAIL (buf, end, 14 * n_entries, need_data);

  g_print ("  %4s|%8s|%4s|%12s|%9s|%12s|%8s\n",
      "N", "unknown", "type", "play item id",
      "time mark", "entry ES PID", "unknown");
  for (i = 0; i < n_entries; i++) {
    guint8 unknown;
    guint8 mark_type;
    guint16 play_item_id;
    guint32 time_mark;
    guint16 entry_es_pid;
    guint32 unknown2;

    unknown = *buf++;
    mark_type = *buf++;
    play_item_id = GST_READ_UINT16_BE (buf);
    buf += 2;
    time_mark = GST_READ_UINT32_BE (buf);
    buf += 4;
    entry_es_pid = GST_READ_UINT16_BE (buf);
    buf += 2;
    unknown2 = GST_READ_UINT32_BE (buf);
    buf += 4;
    g_print ("  %4x|%8d|%4x|%12d|%9d|        %04x|%08x\n",
        i, unknown, mark_type, play_item_id, time_mark, entry_es_pid, unknown2);
  }

  return TRUE;
need_data:
  CHECK_PRINT_ERROR ("Failed to parse mark table");
  return FALSE;
}

static gboolean
mpls_dump_extension_table (guint8 * buf, guint8 * end, guint32 start_offset)
{
  CHECK_AVAIL_DECL;
  guint32 table_size;

  CHECK_AVAIL (buf, end, 4, need_data);
  table_size = GST_READ_UINT32_BE (buf);
  buf += 4;

  g_print ("=== Extension table, size %u bytes (offset 0x%08x-0x%08x) ===\n",
      table_size, start_offset, start_offset + table_size);
  dump_bytes (buf + start_offset, end - buf, 2);

  return TRUE;
need_data:
  CHECK_PRINT_ERROR ("Failed to parse extension table");
  return FALSE;
}

static gboolean
dump_clpi (guint8 * buf, guint32 len)
{
  guint8 *pos, *end;
  guint32 file_id;
  guint32 file_ver;
  guint32 play_items_table_offset;
  guint32 mark_table_offset;
  guint32 extension_table_offset;
  const guint32 bd_version = GST_READ_UINT32_BE ("0200");
  const guint32 avchd_version = GST_READ_UINT32_BE ("0100");

  pos = buf;
  end = pos + len;

  /* Need at least 28 bytes */
  if (pos + 28 > end) {
    g_print ("Not enough bytes to be a CLPI file\n");
    return TRUE;
  }

  file_id = GST_READ_UINT32_BE (pos);
  file_ver = GST_READ_UINT32_BE (pos + 4);
  pos += 8;

  if (file_id != GST_READ_UINT32_BE ("MPLS")) {
    g_print ("Invalid file ID: %" GST_FOURCC_FORMAT ". Should be MPLS\n",
        (guint8) (file_id >> 24), (guint8) (file_id >> 16),
        (guint8) (file_id >> 8), (guint8) (file_id));
    return TRUE;
  }
  if (file_ver != bd_version && file_ver != avchd_version) {
    g_print ("Previously unseen file version: %" GST_FOURCC_FORMAT
        ". Please send it in for analysis\n",
        (guint8) (file_ver >> 24), (guint8) (file_ver >> 16),
        (guint8) (file_ver >> 8), (guint8) (file_ver));
    return TRUE;
  }

  play_items_table_offset = GST_READ_UINT32_BE (pos);
  mark_table_offset = GST_READ_UINT32_BE (pos + 4);
  extension_table_offset = GST_READ_UINT32_BE (pos + 8);
  pos += 12;

  g_print ("Read file header. Play Items table @ 0x%08x Mark table 0x%08x\n",
      play_items_table_offset, mark_table_offset);
  g_print ("    Extension table @ 0x%08x\n", extension_table_offset);

  if (play_items_table_offset == 0) {
    g_print ("Play items table offset == 0 - play items table is mandatory\n");
    return FALSE;
  }

  if (play_items_table_offset >= len) {
    g_print
        ("Not enough data (%d bytes), or play item table offset (%d) invalid\n",
        len, play_items_table_offset);
    return FALSE;
  }
  if (!mpls_dump_play_items_table (buf + play_items_table_offset, end,
          play_items_table_offset))
    return FALSE;

  if (mark_table_offset != 0) {
    if (mark_table_offset >= len) {
      g_print
          ("Not enough data (%d bytes), or mark table offset (%d) invalid\n",
          len, mark_table_offset);
      return FALSE;
    }
    if (!mpls_dump_mark_table (buf + mark_table_offset, end, mark_table_offset))
      return FALSE;
  }

  if (extension_table_offset != 0) {
    if (extension_table_offset >= len) {
      g_print ("Not enough data (%d bytes), or extension table offset "
          "(%d) invalid\n", len, extension_table_offset);
      return FALSE;
    }
    if (!mpls_dump_extension_table (buf + extension_table_offset, end,
            extension_table_offset))
      return FALSE;
  }

  return TRUE;
}

int
main (int argc, char **argv)
{
  int fd, eof = 0;
  int ret = 0;
  guint8 *buf = NULL;
  gchar *contents;
  gsize size;
  GError *error = NULL;
  guint32 avail = 0;

  if (argc < 2) {
    g_print ("Usage: %s <filename>\n", argv[0]);
    return 1;
  }

  g_file_get_contents (argv[1], &contents, &size, &error);
  if (error != NULL) {
    g_print ("Failed opening file %s. Error %s\n", argv[1], error->message);
    g_error_free (error);
    return 1;
  }

  buf = (guint8 *) contents;
  dump_clpi (buf, size);

done:
  g_free (contents);
  close (fd);

  return (ret < 0) ? ret : 0;
}
